import time
import copy
import random
import operator

import torch
import numpy as np
from torch.utils.data import DataLoader
from torch.cuda.amp import GradScaler

from lightwood.logger import log
from lightwood.encoders import NumericEncoder
from lightwood.mixers import BaseMixer
from lightwood.config.config import CONFIG
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.constants.lightwood import COLUMN_DATA_TYPES
from lightwood.mixers.helpers.default_net import DefaultNet
from lightwood.mixers.helpers.selfaware import SelfAware
from lightwood.mixers.helpers.ranger import Ranger
from lightwood.mixers.helpers.transform_corss_entropy_loss import TransformCrossEntropyLoss


class NnMixer(BaseMixer):

    def __init__(self,
                 selfaware=False,
                 callback_on_iter=None,
                 eval_every_x_epochs=20,
                 dropout_p=0.0,
                 stop_training_after_seconds=None,
                 stop_model_building_after_seconds=None,
                 param_optimizer=None):
        """
        :param selfaware: bool
        :param callback_on_iter: Callable[epoch, training_error, test_error, delta_mean, accuracy]
        :param eval_every_x_epochs: int
        :param stop_training_after_seconds: int
        :param stop_model_building_after_seconds: int
        :param param_optimizer: ?
        """
        super().__init__()

        self.selfaware = selfaware
        self.eval_every_x_epochs = eval_every_x_epochs
        self.stop_training_after_seconds = stop_training_after_seconds
        self.stop_model_building_after_seconds = stop_model_building_after_seconds
        self.param_optimizer = param_optimizer

        self.net = None
        self.selfaware_net = None
        self.optimizer = None
        self.selfaware_optimizer = None
        self.optimizer_class = None
        self.optimizer_args = None
        self.selfaware_optimizer_args = None
        self.criterion_arr = None
        self.unreduced_criterion_arr = None

        self.batch_size = 200

        self.nn_class = DefaultNet
        self.dropout_p = max(0.0, min(1.0, dropout_p))
        self.dynamic_parameters = {}
        self.awareness_criterion = None
        self.awareness_scale_factor = 1/10  # scales self-aware total loss contribution
        self.selfaware_lr_factor = 1/2      # scales self-aware learning rate compared to mixer
        self.start_selfaware_training = False
        self.stop_selfaware_training = False
        self.is_selfaware = False

        self.max_confidence_per_output = []
        self.monitor = None

        for k in CONFIG.MONITORING:
            if CONFIG.MONITORING[k]:
                from lightwood.mixers.helpers.debugging import TrainingMonitor
                self.monitor = TrainingMonitor()
                break

        self.total_iterations = 0
        self._nonpersistent = {'sampler': None, 'callback': callback_on_iter}

    def _default_on_iter(self, epoch, train_error, test_error, delta_mean, accuracy):
        test_error_rounded = round(test_error, 4)
        for col in accuracy:
            value = accuracy[col]['value']
            if accuracy[col]['function'] == 'r2_score':
                value_rounded = round(value, 3)
                log.info(f'We\'ve reached training epoch nr {epoch} with an r2 score of {value_rounded} on the testing dataset')
            else:
                value_pct = round(value * 100, 2)
                log.info(f'We\'ve reached training epoch nr {epoch} with an accuracy of {value_pct}% on the testing dataset')

    def _build_confidence_normalization_data(self, ds, subset_id=None):
        self.net = self.net.eval()

        ds.encoders = self.encoders
        ds.transformer = self.transformer

        data_loader = DataLoader(ds, batch_size=self.batch_size, shuffle=True, num_workers=0)

        loss_confidence_arr = []

        for i, data in enumerate(data_loader, 0):
            inputs, labels = data
            inputs = inputs.to(self.net.device)
            labels = labels.to(self.net.device)

            with torch.no_grad():
                if self.is_selfaware:
                    self.selfaware_net.eval()
                    outputs = self.net(inputs)
                    awareness = self.selfaware_net(inputs, outputs)
                else:
                    outputs = self.net(inputs)

            loss = None
            for k, criterion in enumerate(self.criterion_arr):
                if len(loss_confidence_arr) <= k:
                    loss_confidence_arr.append([])
                    self.max_confidence_per_output.append(None)

                try:
                    confidences = criterion.estimate_confidence(outputs[:,ds.out_indexes[k][0]:ds.out_indexes[k][1]])
                    loss_confidence_arr[k].extend(confidences)
                except Exception:
                    pass

        for k, _ in enumerate(self.criterion_arr):
            if len(loss_confidence_arr[k]) > 0:
                loss_confidence_arr[k] = np.array(loss_confidence_arr[k])
                nf_pct = np.percentile(loss_confidence_arr[k], 95)

                losses_bellow_95th_percentile = loss_confidence_arr[k][loss_confidence_arr[k] < nf_pct]
                if len(losses_bellow_95th_percentile) < 1:
                    losses_bellow_95th_percentile = loss_confidence_arr[k]

                self.max_confidence_per_output[k] = max(losses_bellow_95th_percentile)

        return True

    def _fit(self, train_ds, test_ds, stop_training_after_seconds=None):
        """
        :param stop_training_after_seconds: int
        """
        if stop_training_after_seconds is None:
            stop_training_after_seconds = self.stop_training_after_seconds

        input_sample, output_sample = train_ds[0]

        self.net = self.nn_class(
            self.dynamic_parameters,
            input_size=len(input_sample),
            output_size=len(output_sample),
            nr_outputs=len(train_ds.out_types),
            dropout=self.dropout_p
        )
        self.net = self.net.train()

        self.selfaware_net = SelfAware(
            input_size=len(input_sample),
            output_size=len(output_sample),
            nr_outputs=len(train_ds.out_types)
        )
        self.selfaware_net = self.selfaware_net.train()

        if self.batch_size < self.net.available_devices:
            self.batch_size = self.net.available_devices

        self.awareness_criterion = torch.nn.MSELoss()

        if self.criterion_arr is None:
            self.criterion_arr = []
            self.unreduced_criterion_arr = []
            if train_ds.output_weights is not None and train_ds.output_weights is not False:
                output_weights = torch.Tensor(train_ds.output_weights).to(self.net.device)
            else:
                output_weights = None

            for k, output_type in enumerate(train_ds.out_types):
                if output_type in (COLUMN_DATA_TYPES.CATEGORICAL, COLUMN_DATA_TYPES.MULTIPLE_CATEGORICAL):
                    if output_weights is None:
                        weights_slice = None
                    else:
                        # account for numerical features, not included in the output_weights
                        s_idx = train_ds.out_indexes[k][0] - train_ds.output_weights_offset[k]
                        e_idx = train_ds.out_indexes[k][1] - train_ds.output_weights_offset[k]
                        weights_slice = output_weights[s_idx:e_idx]

                    if output_type == COLUMN_DATA_TYPES.CATEGORICAL:
                        self.criterion_arr.append(TransformCrossEntropyLoss(weight=weights_slice))
                        self.unreduced_criterion_arr.append(TransformCrossEntropyLoss(weight=weights_slice, reduction='none'))
                    elif output_type == COLUMN_DATA_TYPES.MULTIPLE_CATEGORICAL:
                        self.criterion_arr.append(torch.nn.BCEWithLogitsLoss(weight=weights_slice))
                        self.unreduced_criterion_arr.append(torch.nn.BCEWithLogitsLoss(weight=weights_slice, reduction='none'))
                # Note: MSELoss works great for numeric, for the other types it's more of a placeholder
                else:
                    self.criterion_arr.append(torch.nn.MSELoss())
                    self.unreduced_criterion_arr.append(torch.nn.MSELoss(reduction='none'))

        self.optimizer_class = Ranger
        if self.optimizer_args is None:
            self.optimizer_args = {'lr': 0.0005}

        if 'beta1' in self.dynamic_parameters:
            self.optimizer_args['betas'] = (self.dynamic_parameters['beta1'], 0.999)

        for optimizer_arg_name in ['lr', 'k', 'N_sma_threshold']:
            if optimizer_arg_name in self.dynamic_parameters:
                self.optimizer_args[optimizer_arg_name] = self.dynamic_parameters[optimizer_arg_name]

        self.optimizer = self.optimizer_class(self.net.parameters(), **self.optimizer_args)

        self.selfaware_optimizer_args = copy.deepcopy(self.optimizer_args)
        self.selfaware_optimizer_args['lr'] = self.selfaware_optimizer_args['lr'] * self.selfaware_lr_factor
        self.selfaware_optimizer = self.optimizer_class(self.selfaware_net.parameters(), **self.optimizer_args)

        if stop_training_after_seconds is None:
            stop_training_after_seconds = round(train_ds.data_frame.shape[0] * train_ds.data_frame.shape[1] / 5)

        if self.stop_model_building_after_seconds is None:
            self.stop_model_building_after_seconds = stop_training_after_seconds * 3

        if self.param_optimizer is not None:
            input_size = len(train_ds[0][0])
            training_data_length = len(train_ds)
            while True:
                training_time_per_iteration = self.stop_model_building_after_seconds / self.param_optimizer.total_trials

                # Some heuristics...
                if training_time_per_iteration > input_size:
                    if training_time_per_iteration > min((training_data_length / (4 * input_size)), 16 * input_size):
                        break

                self.param_optimizer.total_trials = self.param_optimizer.total_trials - 1
                if self.param_optimizer.total_trials < 8:
                    self.param_optimizer.total_trials = 8
                    break

            training_time_per_iteration = self.stop_model_building_after_seconds / self.param_optimizer.total_trials

            self.dynamic_parameters = self.param_optimizer.evaluate(lambda dynamic_parameters: self.evaluate(train_ds, test_ds, dynamic_parameters, max_training_time=training_time_per_iteration, max_epochs=None))

            log.info('Using hyperparameter set: ', self.dynamic_parameters)
        else:
            self.dynamic_parameters = {}

        started = time.time()
        log_reasure = time.time()
        stop_training = False

        for subset_iteration in [1, 2]:
            if stop_training:
                break
            subset_id_arr =  [*train_ds.subsets.keys()]
            for subset_id in subset_id_arr:
                started_subset = time.time()
                if stop_training:
                    break

                subset_train_ds = train_ds.subsets[subset_id]
                subset_test_ds = test_ds.subsets[subset_id]

                lowest_error = None
                last_test_error = None
                last_subset_test_error = None
                test_error_delta_buff = []
                subset_test_error_delta_buff = []
                best_model = None

                #iterate over the iter_fit and see what the epoch and mixer error is
                for epoch, training_error in enumerate(self._iter_fit(subset_train_ds, subset_id=subset_id)):

                    # Log this every now and then so that the user knows it's running
                    if (int(time.time()) - log_reasure) > 30:
                        log_reasure = time.time()
                        log.info(f'Lightwood training, iteration {epoch}, training error {training_error}')


                    # Prime the model on each subset for a bit
                    if subset_iteration == 1:
                        break

                    # Once the training error is getting smaller, enable dropout to teach the network to predict without certain features
                    if subset_iteration > 1 and training_error < 0.4 and not train_ds.enable_dropout:
                        self.eval_every_x_epochs = max(1, int(self.eval_every_x_epochs / 2) )
                        log.info('Enabled dropout !')
                        train_ds.enable_dropout = True
                        lowest_error = None
                        last_test_error = None
                        last_subset_test_error = None
                        test_error_delta_buff = []
                        subset_test_error_delta_buff = []
                        continue

                    # If the selfaware network isn't able to train, go back to the original network
                    if subset_iteration > 1 and (np.isnan(training_error) or np.isinf(training_error) or training_error > pow(10,5)) and not self.stop_selfaware_training:
                        self.start_selfaware_training = False
                        self.stop_selfaware_training = True
                        lowest_error = None
                        last_test_error = None
                        last_subset_test_error = None
                        test_error_delta_buff = []
                        subset_test_error_delta_buff = []
                        continue

                    # Once we are past the priming/warmup period, start training the selfaware network
                    if subset_iteration > 1 and not self.is_selfaware and self.selfaware and not self.stop_selfaware_training and training_error < 0.35:
                        log.info('Started selfaware training !')
                        self.start_selfaware_training = True
                        lowest_error = None
                        last_test_error = None
                        last_subset_test_error = None
                        test_error_delta_buff = []
                        subset_test_error_delta_buff = []
                        continue

                    if epoch % self.eval_every_x_epochs == 0:
                        test_error = self._error(test_ds)
                        subset_test_error = self._error(subset_test_ds, subset_id=subset_id)
                        log.info(f'Subtest test error: {subset_test_error} on subset {subset_id}, overall test error: {test_error}')

                        if lowest_error is None or test_error < lowest_error:
                            lowest_error = test_error
                            best_model = self._get_model_copy()

                        if last_subset_test_error is None:
                            pass
                        else:
                            subset_test_error_delta_buff.append(last_subset_test_error - subset_test_error)

                        last_subset_test_error = subset_test_error

                        if last_test_error is None:
                            pass
                        else:
                            test_error_delta_buff.append(last_test_error - test_error)

                        last_test_error = test_error

                        delta_mean = np.mean(test_error_delta_buff[-5:]) if test_error_delta_buff else 0
                        subset_delta_mean = np.mean(subset_test_error_delta_buff[-5:]) if subset_test_error_delta_buff else 0

                        if self._nonpersistent['callback'] is not None:
                            self._nonpersistent['callback'](epoch, training_error, test_error, delta_mean, self.calculate_accuracy(test_ds))
                        else:
                            self._default_on_iter(epoch, training_error, test_error, delta_mean, self.calculate_accuracy(test_ds))
                        self.net.train()

                        # Stop if we're past the time limit allocated for training
                        if (time.time() - started) > stop_training_after_seconds:
                            stop_training = True

                        # If the trauining subset is overfitting on it's associated testing subset
                        if (subset_delta_mean <= 0 and len(subset_test_error_delta_buff) > 4) or (time.time() - started_subset) > stop_training_after_seconds/len(train_ds.subsets.keys()) or subset_test_error < 0.001:
                            log.info('Finished fitting on {subset_id} of {no_subsets} subset'.format(subset_id=subset_id, no_subsets=len(train_ds.subsets.keys())))

                            self._update_model(best_model)


                            if subset_id == subset_id_arr[-1]:
                                stop_training = True
                            elif not stop_training:
                                break

                        if stop_training:
                            self._update_model(best_model)
                            if self.is_selfaware:
                                self._build_confidence_normalization_data(train_ds)
                                self._adjust(test_ds)

                            self._iter_fit(test_ds, max_epochs=1)
                            self.encoders = train_ds.encoders
                            log.info('Finished training model !')
                            break

    def _predict(self, when_data_source, include_extra_data=False):
        data_loader = DataLoader(
            when_data_source,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=0
        )

        # set model into evaluation mode in order to skip things such as Dropout
        self.net = self.net.eval()

        outputs = []
        awareness_arr = []
        loss_confidence_arr = [[] for _ in when_data_source.out_indexes]

        for i, data in enumerate(data_loader, 0):
            inputs, _ = data
            inputs = inputs.to(self.net.device)

            with torch.no_grad():
                if self.is_selfaware:
                    self.selfaware_net.eval()
                    output = self.net(inputs)
                    awareness = self.selfaware_net(inputs, output)
                    awareness = awareness.to('cpu')
                    awareness_arr.extend(awareness.tolist())
                else:
                    output = self.net(inputs)
                    awareness_arr = None

                for k, criterion in enumerate(self.criterion_arr):
                    try:
                        max_conf = 1
                        if len(self.max_confidence_per_output) >= (k - 1) and self.max_confidence_per_output[k] is not None:
                            max_conf = self.max_confidence_per_output[k]

                        confidences = criterion.estimate_confidence(output[:,when_data_source.out_indexes[k][0]:when_data_source.out_indexes[k][1]], max_conf)
                        loss_confidence_arr[k].extend(confidences)
                    except Exception:
                        loss_confidence_arr[k] = None

                output = output.to('cpu')

            outputs.extend(output)

        output_trasnformed_vectors = {}
        confidence_trasnformed_vectors = {}

        for i in range(len(outputs)):
            output_vector = outputs[i]
            transformed_output_vectors = when_data_source.transformer.revert(
                output_vector, feature_set='output_features')
            for feature in transformed_output_vectors:
                if feature not in output_trasnformed_vectors:
                    output_trasnformed_vectors[feature] = []
                output_trasnformed_vectors[feature].append(transformed_output_vectors[feature])

        predictions = {}
        for k, output_column in enumerate(list(output_trasnformed_vectors.keys())):
            decoded_predictions = when_data_source.get_decoded_column_data(
                output_column,
                torch.Tensor(output_trasnformed_vectors[output_column])
            )

            predictions[output_column] = {'predictions': decoded_predictions['predictions']}

            if awareness_arr is not None:
                predictions[output_column]['selfaware_confidences'] = [1/abs(x[k]) if x[k] != 0 else 1/0.000001 for x in awareness_arr]

            if loss_confidence_arr[k] is not None:
                predictions[output_column]['loss_confidences'] = loss_confidence_arr[k]

            if include_extra_data:
                predictions[output_column]['encoded_predictions'] = output_trasnformed_vectors[output_column]

            if 'class_distribution' in decoded_predictions:
                predictions[output_column]['class_distribution'] = decoded_predictions['class_distribution']
                predictions[output_column]['class_labels'] = decoded_predictions['class_labels']

        log.info('Model predictions and decoding completed')

        return predictions

    def overall_certainty(self):
        """
        return an estimate of how certain is the model about the overall predictions,
        in this case its a measurement of how much did the variance of all the weights distributions
        reduced from its initial distribution
        :return:
        """
        if hasattr(self.net, 'calculate_overall_certainty'):
            return self.net.calculate_overall_certainty()
        else:
            return -1

    def _error(self, ds, subset_id=None):
        self.net = self.net.eval()

        ds.encoders = self.encoders
        ds.transformer = self.transformer

        if self._nonpersistent['sampler'] is None:
            data_loader = DataLoader(ds, batch_size=self.batch_size, sampler=self._nonpersistent['sampler'], num_workers=0)
        else:
            data_loader = DataLoader(ds, batch_size=self.batch_size, shuffle=True, num_workers=0)

        running_loss = 0.0
        error = 0

        for i, data in enumerate(data_loader, 0):
            inputs, labels = data
            inputs = inputs.to(self.net.device)
            labels = labels.to(self.net.device)

            with torch.no_grad():
                outputs = self.net(inputs)

            loss = None
            for k, criterion in enumerate(self.criterion_arr):
                target_loss = criterion(outputs[:,ds.out_indexes[k][0]:ds.out_indexes[k][1]], labels[:,ds.out_indexes[k][0]:ds.out_indexes[k][1]])

                if loss is None:
                    loss = target_loss
                else:
                    loss += target_loss

            running_loss += loss.item()
            error = running_loss / (i + 1)

        if CONFIG.MONITORING['epoch_loss']:
            self.monitor.plot_loss(error, self.total_iterations, 'Test Epoch Error')
            self.monitor.plot_loss(error, self.total_iterations, f'Test Epoch Error - Subset {subset_id}')

        self.net = self.net.train()

        return error

    def _get_model_copy(self):
        """
        get the actual mixer model
        :return: self.net
        """
        self.optimizer.zero_grad()
        return copy.deepcopy(self.net)

    def _update_model(self, model):
        """
        replace the current model with a model object
        :param model: a model object
        :return: None
        """

        self.net = model

        if 'cuda' in str(self.net.device):
            torch.cuda.empty_cache()
        self.optimizer.zero_grad()
        self.optimizer = self.optimizer_class(self.net.parameters(), **self.optimizer_args)

    def _iter_fit(self, ds, subset_id=None, max_epochs=120000):
        if self._nonpersistent['sampler'] is None:
            data_loader = DataLoader(
                ds,
                batch_size=self.batch_size,
                shuffle=True,
                num_workers=0
            )
        else:
            data_loader = DataLoader(
                ds,
                batch_size=self.batch_size,
                num_workers=0,
                sampler=self._nonpersistent['sampler']
            )

        scaler = GradScaler()
        selfaware_scaler = GradScaler()

        for epoch in range(max_epochs):  # loop over the dataset multiple times
            running_loss = 0.0
            error = 0
            for i, data in enumerate(data_loader, 0):
                if self.start_selfaware_training and not self.is_selfaware:
                    log.info('Starting to train selfaware network for better confidence determination !')
                    self.is_selfaware = True

                if self.stop_selfaware_training and self.is_selfaware:
                    log.info('Cannot train selfaware network, will fallback to using simpler confidence models !')
                    self.is_selfaware = False

                self.total_iterations += 1
                # get the inputs; data is a list of [inputs, labels]
                inputs, labels = data

                labels = labels.to(self.net.device)
                inputs = inputs.to(self.net.device)

                # zero the parameter gradients
                self.optimizer.zero_grad()
                self.selfaware_optimizer.zero_grad()

                # forward + backward + optimize
                with LightwoodAutocast():
                    outputs = self.net(inputs)
                if self.is_selfaware:
                    with LightwoodAutocast():
                        awareness = self.selfaware_net(inputs.detach(), outputs.detach())

                loss = None
                for k, criterion in enumerate(self.criterion_arr):
                    with LightwoodAutocast():
                        target_loss = criterion(outputs[:, ds.out_indexes[k][0]:ds.out_indexes[k][1]],
                                                labels[:, ds.out_indexes[k][0]:ds.out_indexes[k][1]])

                    if loss is None:
                        loss = target_loss
                    else:
                        loss += target_loss

                awareness_loss = None
                if self.is_selfaware:
                    unreduced_losses = []
                    for k, criterion in enumerate(self.unreduced_criterion_arr):
                        target_loss = criterion(
                            outputs[:,ds.out_indexes[k][0]:ds.out_indexes[k][1]],
                            labels[:,ds.out_indexes[k][0]:ds.out_indexes[k][1]]
                        )

                        target_loss = target_loss.tolist()
                        if type(target_loss[0]) == type([]):
                            target_loss = [np.mean(x) for x in target_loss]
                        for j, value in enumerate(target_loss):
                            if len(unreduced_losses) <= j:
                                unreduced_losses.append([])
                            unreduced_losses[j].append(value)

                    unreduced_losses = torch.Tensor(unreduced_losses).to(self.net.device)

                    awareness_loss = self.awareness_criterion(awareness, unreduced_losses) * self.awareness_scale_factor

                    if CONFIG.MONITORING['batch_loss']:
                        self.monitor.plot_loss(awareness_loss.item(), self.total_iterations, 'Awreness Batch Loss')

                if CONFIG.MONITORING['batch_loss']:
                    self.monitor.plot_loss(loss.item(), self.total_iterations, 'Targets Batch Loss')

                running_loss += loss.item()

                if LightwoodAutocast.active:
                    scaler.scale(loss).backward()
                    scaler.step(self.optimizer)
                    scaler.update()
                    if self.is_selfaware and self.start_selfaware_training and awareness_loss is not None:
                        selfaware_scaler.scale(awareness_loss).backward()
                        selfaware_scaler.step(self.selfaware_optimizer)
                        selfaware_scaler.update()
                else:
                    loss.backward()
                    self.optimizer.step()
                    if self.is_selfaware and self.start_selfaware_training and awareness_loss is not None:
                        awareness_loss.backward(retain_graph=True)
                        self.selfaware_optimizer.step()

                error = running_loss / (i + 1)

                # @NOTE: Decrease 900 if you want to plot gradients more often, I find it's too expensive to do so
                if CONFIG.MONITORING['network_heatmap'] and random.randint(0, 1000) > 900:
                    weights = []
                    gradients = []
                    layer_name = []
                    for index, layer in enumerate(self.net.net):
                        if 'Linear' in str(type(layer)):
                            weights.append( list(layer.weight.cpu().detach().numpy().ravel()) )
                            gradients.append( list(layer.weight.grad.cpu().detach().numpy().ravel()) )
                            layer_name.append(f'Layer {index}-{index+1}')
                    self.monitor.weight_map(layer_name, weights, 'Predcitive network weights')
                    self.monitor.weight_map(layer_name, weights, 'Predictive network gradients')

                    if self.is_selfaware:
                        weights = []
                        gradients = []
                        layer_name = []
                        for index, layer in enumerate(self.selfaware_net.net):
                            if 'Linear' in str(type(layer)):
                                weights.append( list(layer.weight.cpu().detach().numpy().ravel()) )
                                gradients.append( list(layer.weight.grad.cpu().detach().numpy().ravel()) )
                                layer_name.append(f'Layer {index}-{index+1}')
                        self.monitor.weight_map(layer_name, weights, 'Awareness network weights')
                        self.monitor.weight_map(layer_name, weights, 'Awareness network gradients')

                if CONFIG.MONITORING['batch_loss']:
                    #self.monitor.plot_loss(total_loss.item(), self.total_iterations, 'Total Batch Loss')
                    self.monitor.plot_loss(error, self.total_iterations, 'Mean Total Running Loss')

            if CONFIG.MONITORING['epoch_loss']:
                self.monitor.plot_loss(error, self.total_iterations, 'Train Epoch Error')
                self.monitor.plot_loss(error, self.total_iterations, f'Train Epoch Error - Subset {subset_id}')
            yield error

    def to(self, device, available_devices):
        self.net.to(device, available_devices)
        self.selfaware_net.to(device, available_devices)
        for enc in self.encoders:
            self.encoders[enc].to(device, available_devices)
        return self

    def calculate_accuracy(self, ds):
        predictions = self.predict(ds, include_extra_data=True)
        accuracies = {}

        for output_column in [feature['name'] for feature in ds.config['output_features']]:

            col_type = ds.get_column_config(output_column)['type']

            if col_type == COLUMN_DATA_TYPES.MULTIPLE_CATEGORICAL:
                reals = [tuple(x) for x in ds.get_column_original_data(output_column)]
                preds = [tuple(x) for x in predictions[output_column]['predictions']]
            else:
                reals = [str(x) for x in ds.get_column_original_data(output_column)]
                preds = [str(x) for x in predictions[output_column]['predictions']]

            if 'weights' in ds.get_column_config(output_column):
                weight_map = ds.get_column_config(output_column)['weights']
                # omit points for which there's no target info in timeseries forecasting
                if '_timestep_' in ds.get_column_config(output_column)['name']:
                    classes = set(weight_map.keys())
                    observed_classes = set(reals)
                    for unrecognized_class in observed_classes.difference(classes):
                        weight_map[unrecognized_class] = 0
            else:
                weight_map = None

            accuracy = BaseMixer._apply_accuracy_function(
                ds.get_column_config(output_column)['type'],
                reals,
                preds,
                weight_map=weight_map,
                encoder=ds.encoders[output_column]
            )

            if ds.get_column_config(output_column)['type'] == COLUMN_DATA_TYPES.NUMERIC:
                if isinstance(ds.encoders[output_column], NumericEncoder):
                    ds.encoders[output_column].decode_log = True
                else:
                    ds.encoders[output_column].decode_log = False  # TsNumericEncoder case
                preds = ds.get_decoded_column_data(
                    output_column,
                    predictions[output_column]['encoded_predictions']
                )['predictions']

                alternative_accuracy = BaseMixer._apply_accuracy_function(
                    ds.get_column_config(output_column)['type'],
                    reals,
                    preds,
                    weight_map=weight_map
                )

                if alternative_accuracy['value'] > accuracy['value']:
                    accuracy = alternative_accuracy
                else:
                    ds.encoders[output_column].decode_log = False

            accuracies[output_column] = accuracy

        return accuracies



import importlib
import inspect
import copy
import random
import string

import torch
import numpy as np
import pandas as pd
from torch.utils.data import Dataset

from lightwood.mixers.helpers.transformer import Transformer
from lightwood.encoders.time_series.helpers.common import generate_target_group_normalizers
from lightwood.config.config import CONFIG
from lightwood.constants.lightwood import ColumnDataTypes
from lightwood.encoders import (
    NumericEncoder,
    TsNumericEncoder,
    CategoricalAutoEncoder,
    MultiHotEncoder,
    DistilBertEncoder,
    DatetimeEncoder,
    Img2VecEncoder,
    TsRnnEncoder,
    TimeSeriesPlainEncoder,
    ShortTextEncoder,
    VocabularyEncoder
)


class SubSet(Dataset):
    def __init__(self, data_source, indexes):
        self.data_source = data_source
        self.index_mapping = {}
        for i in range(len(indexes)):
            self.index_mapping[i] = indexes[i]

    def __len__(self):
        return int(len(self.index_mapping.keys()))

    def __getitem__(self, idx):
        return self.data_source[self.index_mapping[idx]]

    def get_feature_names(self, where='input_features'):
        return self.data_source.get_feature_names(where)

    def __getattribute__(self, name):
        if name in ['config', 'encoders', 'transformer', 'training',
                    'output_weights', 'dropout_dict', 'disable_cache', 'out_types', 'out_indexes']:
            return self.data_source.__getattribute__(name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name in ['config', 'encoders', 'transformer', 'training',
                    'output_weights', 'dropout_dict', 'disable_cache']:
            return dict.__setattr__(self.data_source, name, value)
        else:
            super().__setattr__(name, value)


class DataSource(Dataset):
    def __init__(self,
                 data_frame,
                 config,
                 prepare_encoders=True,
                 initialize_transformer=True):
        """
        Create a lightwood datasource from the data frame
        :param data_frame:
        :param config
        """
        self.subsets = {}
        self.data_frame = data_frame
        self.config = config
        self.training = False  # Flip this flag if you are using the datasource while training
        self.output_weights = None
        self.output_weights_offset = None
        self.dropout_dict = {}
        self.enable_dropout = False
        self.enable_cache = self.config['data_source']['cache_transformed_data']
        self.out_indexes = None

        self.out_types = [feature['type'] for feature in self.config['output_features']]
        self.output_feature_names = [feature['name'] for feature in self.config['output_features']]
        self.input_feature_names = [feature['name'] for feature in self.config['input_features']]
        self.output_features = self.config['output_features']
        self.input_features = self.config['input_features']

        for col in self.input_features:
            if len(self.input_features) > 1:
                dropout = 0.1
            else:
                dropout = 0.0

            if 'dropout' in col:
                dropout = col['dropout']

            self.dropout_dict[col['name']] = dropout

        for col in self.config['output_features']:
            prev_col = [c for c in self.config['input_features'] if c['name'] == f'__mdb_ts_previous_{col["name"]}']
            # len(prev_col) <= 1 always holds
            if prev_col:
                col['group_info'] = {
                    prev_col['name']: self.get_column_original_data(prev_col['name'])
                    for prev_col in self.config['input_features'] if prev_col['grouped_by']
                }
        
        if initialize_transformer:
            self.transformer = Transformer(self.input_feature_names, self.output_feature_names)
        else:
            self.transformer = None

        if prepare_encoders:
            self.encoders = self._prepare_encoders()
        else:
            self.encoders = None
        
        self._clear_cache()

    def extend(self, df):
        """
        :df: pandas.DataFrame or DataSource
        """
        if isinstance(df, pd.DataFrame):
            self.data_frame.append(df)
        else:
            raise TypeError(':df: must be pandas.DataFrame')

    def create_subsets(self, nr_subsets):
        subsets_indexes = {}
        np.random.seed(len(self.data_frame))

        subset_nr = 1
        for i in range(len(self.data_frame)):
            if subset_nr not in subsets_indexes:
                subsets_indexes[subset_nr] = []
            subsets_indexes[subset_nr].append(i)

            subset_nr += 1
            if subset_nr > nr_subsets:
                subset_nr = 1

        for subset_nr in subsets_indexes:
            self.subsets[subset_nr] = SubSet(self, subsets_indexes[subset_nr])

    def _clear_cache(self):
        self.encoded_cache = {}
        self.transformed_cache = None

    def subset(self, percentage):
        """
        Removes :percentage: of data randomly from itself and returns a new
        datasource made from that data

        :param percentage: float

        :return: DataSource
        """
        np.random.seed(int(round(percentage * 100000)))

        msk = np.random.rand(len(self.data_frame)) < (1 - percentage)

        sub_df = self.data_frame[~msk]
        self.data_frame = self.data_frame[msk]

        self._clear_cache()

        ds = DataSource(
            sub_df,
            self.config,
            prepare_encoders=False,
            initialize_transformer=False
        )
        ds.encoders = self.encoders
        ds.transformer = self.transformer
        ds.output_weights = self.output_weights

        return ds

    def __len__(self):
        """
        return the length of the datasource (as in number of rows)
        :return: number of rows
        """
        return int(self.data_frame.shape[0])

    def __getitem__(self, idx):
        sample = {}

        dropout_features = None

        if self.training and random.randint(0, 3) == 1 and self.enable_dropout and CONFIG.ENABLE_DROPOUT:
            dropout_features = [feature['name'] for feature in self.config['input_features'] if random.random() > (1 - self.dropout_dict[feature['name']])]

            # Make sure we never drop all the features, since this would make the row meaningless
            if len(dropout_features) > len(self.config['input_features']):
                dropout_features = dropout_features[:-1]
            #logging.debug(f'\n-------------\nDroping out features: {dropout_features}\n-------------\n')

        if self.enable_cache:
            if self.transformed_cache is None:
                self.transformed_cache = [None] * len(self)

            if dropout_features is None or len(dropout_features) == 0:
                cached_sample = self.transformed_cache[idx]
                if cached_sample is not None:
                    return cached_sample

        for feature_set in ['input_features', 'output_features']:
            sample[feature_set] = {}
            for feature in self.config[feature_set]:
                col_name = feature['name']
                col_config = self.get_column_config(col_name)

                if col_name not in self.encoded_cache:  # if data is not encoded yet, encode values
                    if not ((dropout_features is not None and col_name in dropout_features) or not self.enable_cache):
                        self.get_encoded_column_data(col_name)

                # if we are dropping this feature, get the encoded value of None
                if (dropout_features is not None and col_name in dropout_features):
                    custom_data = {col_name: [None]}
                    # if the dropout feature depends on another column, also pass a None array as the dependant column
                    if 'depends_on_column' in col_config:
                        for col in col_config['depends_on_column']:
                            custom_data[col] = [None]
                    sample[feature_set][col_name] = self.get_encoded_column_data(col_name, custom_data=custom_data)[0]
                elif not self.enable_cache:
                    if col_name in self.data_frame:
                        custom_data = {col_name: [self.data_frame[col_name].iloc[idx]]}
                        if 'depends_on_column' in col_config:
                            for col in col_config['depends_on_column']:
                                custom_data[col] = [self.data_frame[col].iloc[idx]]
                    else:
                        custom_data = {col_name: [None]}

                    sample[feature_set][col_name] = self.get_encoded_column_data(col_name, custom_data=custom_data)[0]
                else:
                    sample[feature_set][col_name] = self.encoded_cache[col_name][idx]

        # Create weights if not already created
        if self.output_weights is None:
            for col_config in self.config['output_features']:
                if 'weights' in col_config:

                    weights = col_config['weights']
                    new_weights = None

                    for val in weights:
                        encoded_val = self.get_encoded_column_data(
                            col_config['name'],
                            custom_data={
                                col_config['name']: [val]
                            }
                        )

                        if new_weights is None:
                            mean = np.mean(list(weights.values()))
                            new_weights = [mean] * len(encoded_val[0])

                        for value_index, hot in enumerate(encoded_val[0]):
                            if hot:
                                new_weights[value_index] = weights[val] / sum(encoded_val[0]).item()

                    if self.output_weights is None or self.output_weights == False:
                        self.output_weights = new_weights
                    else:
                        self.output_weights.extend(new_weights)
                else:
                    self.output_weights = False

        sample = self.transformer.transform(sample)
        if self.out_indexes is None:
            self.out_indexes = self.transformer.out_indexes

        if self.enable_cache:
            self.transformed_cache[idx] = sample

        if self.output_weights_offset is None:
            self.output_weights_offset = {}
            for idx, (otype, oidxs) in enumerate(zip(self.out_types, self.out_indexes)):
                self.output_weights_offset[idx] = self.output_weights_offset.get(idx-1, 0)
                if otype not in (ColumnDataTypes.CATEGORICAL, ColumnDataTypes.MULTIPLE_CATEGORICAL):
                     self.output_weights_offset[idx] += (oidxs[1] - oidxs[0])

        return sample

    def get_column_original_data(self, column_name):
        if column_name not in self.data_frame:
            nr_rows = self.data_frame.shape[0]
            return [None] * nr_rows

        return self.data_frame[column_name].tolist()

    def _lookup_encoder_class(self, column_config, is_target):
        default_encoder_classes = {
            ColumnDataTypes.NUMERIC: NumericEncoder,
            ColumnDataTypes.CATEGORICAL: CategoricalAutoEncoder,
            ColumnDataTypes.MULTIPLE_CATEGORICAL: MultiHotEncoder,
            ColumnDataTypes.DATETIME: DatetimeEncoder,
            ColumnDataTypes.IMAGE: Img2VecEncoder,
            ColumnDataTypes.TEXT: DistilBertEncoder,
            ColumnDataTypes.SHORT_TEXT: ShortTextEncoder,
            ColumnDataTypes.TIME_SERIES: TsRnnEncoder,
            # ColumnDataTypes.AUDIO: AmplitudeTsEncoder
        }

        target_encoder_classes = {
            ColumnDataTypes.TEXT: VocabularyEncoder
        }

        column_type = column_config['type']
        if is_target and column_type in target_encoder_classes:
            encoder_class = target_encoder_classes[column_type]
        else:
            encoder_class = default_encoder_classes[column_type]

        # specialized dispatches
        if is_target and column_config.get('additional_info', {}).get('time_series_target', False) and \
                column_type == ColumnDataTypes.NUMERIC:
            encoder_class = TsNumericEncoder
        if column_type == ColumnDataTypes.TIME_SERIES and '__mdb_ts_previous' in column_config['name']:
            encoder_class = TimeSeriesPlainEncoder

        return encoder_class

    def _make_column_encoder(self, encoder_class, encoder_attrs=None, is_target=False):
        encoder_instance = encoder_class(is_target=is_target)
        encoder_attrs = encoder_attrs or {}
        for attr in encoder_attrs:
            if hasattr(encoder_instance, attr):
                setattr(encoder_instance, attr, encoder_attrs[attr])
        return encoder_instance

    def _prepare_column_encoder(self, config, is_target=False, training_data=None):
        column_data = self.get_column_original_data(config['name'])
        encoder_class = config.get(
            'encoder_class',
            self._lookup_encoder_class(config, is_target)
        )
        encoder_attrs = config.get('encoder_attrs', {})
        encoder_attrs['original_type'] = config.get('original_type', None)
        encoder_attrs['secondary_type'] = config.get('secondary_type', None)

        encoder_instance = self._make_column_encoder(
            encoder_class,
            encoder_attrs,
            is_target=is_target
        )

        if training_data and 'training_data' in inspect.getfullargspec(encoder_instance.prepare).args:
            encoder_instance.prepare(
                column_data,
                training_data=training_data
            )
        else:
            # joint column data augmentation for time series
            if config['type'] == ColumnDataTypes.TIME_SERIES and not is_target and '__mdb_ts_previous' not in config['name']:
                encoder_instance.prepare(column_data, previous_target_data=training_data['previous'])
            else:
                encoder_instance.prepare(column_data)

        return encoder_instance

    def _prepare_encoders(self):
        """
        Get the encoder for all the output and input column and prepare them
        with all available data for that column.
        """
        encoders = {}
        input_encoder_training_data = {'targets': [], 'previous': []}

        # handle time series previous target columns
        previous_cols = []
        for config in self.config['input_features']:
            column_name = config['name']
            if column_name.startswith('__mdb_ts_previous_'):
                column_data = self.get_column_original_data(column_name)
                previous_cols.append(column_name)
                col_info = {'data': column_data,
                            'name': column_name,
                            'original_type': config['original_type'],
                            'group_info': {conf['name']: self.get_column_original_data(conf['name'])
                                           for conf in self.config['input_features'] if conf['grouped_by']},
                            'output_type': config['type']
                            }
                col_info = generate_target_group_normalizers(col_info)
                input_encoder_training_data['previous'].append(col_info)
                config['additional_info']['normalizers'] = col_info['normalizers']
                config['additional_info']['group_combinations'] = col_info['group_combinations']

        # train output encoders (except for time-series related)
        for config in self.config['output_features']:
            if config['type'] not in ColumnDataTypes.TIME_SERIES:
                column_name = config['name']
                column_data = self.get_column_original_data(column_name)
                encoder_instance = self._prepare_column_encoder(config, is_target=True)

                input_encoder_training_data['targets'].append({
                    'encoded_output': encoder_instance.encode(column_data),
                    'unencoded_output': column_data,
                    'output_encoder': encoder_instance,
                    'output_type': config['type']
                })

                encoders[column_name] = encoder_instance

        # train input encoders
        for config in self.config['input_features']:
            column_name = config['name']
            encoder_instance = self._prepare_column_encoder(config,
                                                            is_target=False,
                                                            training_data=input_encoder_training_data)
            encoders[column_name] = encoder_instance

            if column_name not in previous_cols:
                if config['type'] == ColumnDataTypes.TIME_SERIES and len(input_encoder_training_data['previous']) > 0:
                    for d in input_encoder_training_data['previous']:
                        try:
                            if not d['name'] in config['depends_on_column']:
                                config['depends_on_column'].append(d['name'])
                        except KeyError:
                            config['depends_on_column'] = [d['name']]

        # train time series output encoder
        for config in self.config['output_features']:
            column_name = config['name']

            prev_col = [c for c in self.config['input_features'] if
                        c['name'] == f'__mdb_ts_previous_{column_name}']
            if prev_col:
                # 0-indexed as we know it can only be one
                config['depends_on_column'] = [prev_col[0]['name']]
                config['encoder_attrs']['normalizers'] = prev_col[0]['additional_info']['normalizers']
                config['encoder_attrs']['group_combinations'] = prev_col[0]['additional_info'][
                    'group_combinations']

            encoder_instance = self._prepare_column_encoder(config, is_target=True)
            encoders[column_name] = encoder_instance

        return encoders

    def make_child(self, df):
        """
        :param df: DataFrame

        :return: DataSource
        """
        child = DataSource(
            df,
            self.config,
            prepare_encoders=False,
            initialize_transformer=False
        )
        child.transformer = self.transformer
        child.encoders = self.encoders
        child.output_weights = self.output_weights
        return child

    def get_encoded_column_data(self, column_name, custom_data=None):
        if column_name in self.encoded_cache and custom_data is None:
            return self.encoded_cache[column_name]

        # The first argument of encoder is the data, if no custom data is specified,
        # use all the datasource's data for this column
        if custom_data is not None:
            args = [custom_data[column_name]]
        else:
            args = [self.get_column_original_data(column_name)]

        config = self.get_column_config(column_name)

        # Pass dependency info when applicable
        encoder_arity = len(inspect.signature(self.encoders[column_name].encode).parameters)
        if encoder_arity >= 2 and 'depends_on_column' in config:
            arg2 = []
            for col in config['depends_on_column']:
                sublist = {'group_info': {conf['name']: self.get_column_original_data(conf['name'])
                                          for conf in self.config['input_features'] if conf['grouped_by']},
                           'name': col}
                if custom_data is None:
                    sublist['data'] = self.get_column_original_data(col)
                else:
                    sublist['data'] = custom_data.get(col, None)
                arg2.append(sublist)
            args.append(arg2)

        encoded_vals = self.encoders[column_name].encode(*args)
        if isinstance(encoded_vals, list) and isinstance(encoded_vals[0], torch.Tensor):
            encoded_vals = torch.stack(encoded_vals)

        # Cache the encoded data so we don't have to run the encoding,
        # Don't cache custom_data
        # (custom_data is usually used when running without cache or dropping out a feature for a certain pass)
        if column_name not in self.encoded_cache and custom_data is None:
            self.encoded_cache[column_name] = encoded_vals
        return encoded_vals

    def get_decoded_column_data(self, column_name, encoded_data, decoder_instance=None):
        """
        :param column_name: column names to be decoded
        :param encoded_data: encoded data of tensor type
        :return decoded_data : Dict :Decoded data of input column
        """
        if decoder_instance is None:
            if column_name not in self.encoders:
                raise ValueError("""
                    Data must have been encoded before at some point.
                    You should not decode before having encoding at least once
                    """)
            decoder_instance = self.encoders[column_name]

        decoded_data = {}
        if getattr(decoder_instance, 'predict_proba', False):
            # return complete belief distribution
            preds, pred_probs, labels = decoder_instance.decode(encoded_data)
            decoded_data['predictions'] = preds
            decoded_data['class_distribution'] = pred_probs
            decoded_data['class_labels'] = labels
        elif isinstance(decoder_instance, TsNumericEncoder):
            group_info = {conf['name']: self.get_column_original_data(conf['name'])
                          for conf in self.config['input_features'] if conf.get('grouped_by', False)}
            decoded_data['predictions'] = decoder_instance.decode(encoded_data, group_info=group_info)
        else:
            decoded_data['predictions'] = decoder_instance.decode(encoded_data)

        return decoded_data

    def get_feature_names(self, where='input_features'):
        return [feature['name'] for feature in self.config[where]]

    def get_column_config(self, column_name):
        """
        Get the config info for the feature given a config as defined in data_schemas definition.py
        :param column_name:
        :return:
        """
        for feature_set in ['input_features', 'output_features']:
            for feature in self.config[feature_set]:
                if feature['name'] == column_name:
                    return feature
        return None

    def train(self):
        self.training = True
    
    def eval(self):
        self.training = False



import os
import torch
from random import randint

from lightwood.config.config import CONFIG


def get_devices():
    if CONFIG.USE_CUDA and torch.cuda.is_available():
        device_str = "cuda"
        available_devices = torch.cuda.device_count()

        if available_devices > 1:
            if os.environ.get('RANDOM_GPU', False) in ['1', 'true', 'True', True, 1]:
                device_str = 'cuda:' + str(randint(0,available_devices-1))
                available_devices = 1
    else:
        device_str = "cpu"
        available_devices = 1

    if CONFIG.USE_DEVICE is not None and os.environ.get('RANDOM_GPU', False) not in ['1', 'true', 'True', True, 1]:
        device_str = CONFIG.USE_DEVICE
        if device_str != 'cuda':
            available_devices = 1

    device = torch.device(device_str)

    return device, available_devices



import numpy as np
import optuna.integration.lightgbm as lgb
import lightgbm
import optuna
import torch
import logging
import time

from lightwood.constants.lightwood import COLUMN_DATA_TYPES
from lightwood.helpers.device import get_devices
from lightwood.mixers import BaseMixer
from sklearn.preprocessing import OrdinalEncoder


optuna.logging.set_verbosity(optuna.logging.CRITICAL)


def check_gpu_support():
    try:
        data = np.random.rand(50, 2)
        label = np.random.randint(2, size=50)
        train_data = lightgbm.Dataset(data, label=label)
        params = {'num_iterations': 1, 'device': 'gpu'}
        gbm = lightgbm.train(params, train_set=train_data)
        return True
    except Exception as e:
        return False

class LightGBMMixer(BaseMixer):
    def __init__(self, stop_training_after_seconds=None, grid_search=False):
        super().__init__()
        self.models = {}
        self.ord_encs = {}
        self.label_sets = {}
        self.stop_training_after_seconds = stop_training_after_seconds
        self.grid_search = grid_search  # using Optuna

        # GPU Only available via --install-option=--gpu with opencl-dev and libboost dev (a bunch of them) installed, so let's turn this off for now and we can put it behind some flag later
        self.device, _ = get_devices()
        self.device_str = 'cpu' if str(self.device) == 'cpu' else 'gpu'
        if self.device_str == 'gpu':
            gpu_works = check_gpu_support()
            if not gpu_works:
                self.device = torch.device('cpu')
                self.device_str = 'cpu'

        self.max_bin = 255  # Default value
        if self.device_str == 'gpu':
            self.max_bin = 63  # As recommended by https://lightgbm.readthedocs.io/en/latest/Parameters.html#device_type

    def _fit(self, train_ds, test_ds=None):
        """
        :param train_ds: DataSource
        :param test_ds: DataSource
        """

        data = {
            'train': {'ds': train_ds, 'data': None, 'label_data': {}},
            'test': {'ds': test_ds, 'data': None, 'label_data': {}}
        }

        # Order is important here
        for subset_name in ['train','test']:
            cols = data[subset_name]['ds'].input_feature_names
            out_cols = data[subset_name]['ds'].output_feature_names
            for col_name in cols:
                if data[subset_name]['data'] is None:
                    data[subset_name]['data'] = data[subset_name]['ds'].get_encoded_column_data(col_name).to(self.device)
                else:
                    enc_col = data[subset_name]['ds'].get_encoded_column_data(col_name)
                    data[subset_name]['data'] = torch.cat((data[subset_name]['data'], enc_col.to(self.device)), 1)
            data[subset_name]['data'] = data[subset_name]['data'].tolist()
            for col_name in out_cols:
                label_data = data[subset_name]['ds'].get_column_original_data(col_name)
                if next(item for item in train_ds.output_features if item["name"] == col_name)['type'] == COLUMN_DATA_TYPES.CATEGORICAL:
                    if subset_name == 'train':
                        self.ord_encs[col_name] = OrdinalEncoder()
                        self.label_sets[col_name] = set(label_data)
                        self.label_sets[col_name].add('__mdb_unknown_cat')
                        self.ord_encs[col_name].fit(np.array(list(self.label_sets[col_name])).reshape(-1, 1))

                    label_data = [x if x in self.label_sets[col_name] else '__mdb_unknown_cat' for x in label_data]
                    label_data = self.ord_encs[col_name].transform(np.array(label_data).reshape(-1, 1)).flatten()

                data[subset_name]['label_data'][col_name] = label_data

        for col_name in train_ds.output_feature_names:
            dtype = next(item for item in train_ds.output_features if item["name"] == col_name)['type']
            if dtype not in [COLUMN_DATA_TYPES.NUMERIC, COLUMN_DATA_TYPES.CATEGORICAL]:
                logging.info('cannot support {dtype} in lightgbm'.format(dtype=dtype))
                continue
            else:
                objective = 'regression' if dtype == COLUMN_DATA_TYPES.NUMERIC else 'multiclass'
                metric = 'l2' if dtype == COLUMN_DATA_TYPES.NUMERIC else 'multi_logloss'

            params = {'objective': objective,
                      'metric': metric,
                      'verbose': -1,
                      'lambda_l1': 0.1,
                      'lambda_l2': 0.1,
                      'force_row_wise': True,
                      'device_type': self.device_str
                      }
            if objective == 'multiclass':
                self.all_classes = self.ord_encs[col_name].categories_[0]
                params['num_class'] = self.all_classes.size

            num_iterations = 100

            if self.stop_training_after_seconds is not None:
                train_data = lightgbm.Dataset(data['train']['data'], label=data['train']['label_data'][col_name])
                validate_data = lightgbm.Dataset(data['test']['data'], label=data['test']['label_data'][col_name])
                start = time.time()
                params['num_iterations'] = 1
                bst = lightgbm.train(params, train_data, valid_sets=validate_data, verbose_eval=False)
                end = time.time()
                seconds_for_one_iteration = min(0.1, end - start)
                logging.info(f'A single GBM itteration takes {seconds_for_one_iteration} seconds')
                max_itt = int(self.stop_training_after_seconds/seconds_for_one_iteration)
                num_iterations = max(1, min(num_iterations, max_itt))
                # Turn on grid search if training doesn't take too long using it
                if max_itt > 10*num_iterations and seconds_for_one_iteration < 10:
                    self.grid_search = True

            train_data = lightgbm.Dataset(data['train']['data'], label=data['train']['label_data'][col_name])
            validate_data = lightgbm.Dataset(data['test']['data'], label=data['test']['label_data'][col_name])
            model = lgb if self.grid_search else lightgbm
            logging.info(f'Training GBM ({model}) with {num_iterations} iterations')
            params['num_iterations'] = num_iterations
            bst = model.train(params, train_data, valid_sets=validate_data, verbose_eval=False)

            self.models[col_name] = bst

    def _predict(self, when_data_source, include_extra_data=False):
        """
        :param when_data_source: DataSource
        :param include_extra_data: bool
        """
        data = None
        for col_name in when_data_source.input_feature_names:
            if data is None:
                data = when_data_source.get_encoded_column_data(col_name).to(self.device)
            else:
                data = torch.cat((data, when_data_source.get_encoded_column_data(col_name).to(self.device)), 1)
        data = data.tolist()

        ypred = {}
        for col_name in when_data_source.output_feature_names:
            col_config = [conf for conf in when_data_source.config['output_features'] if conf['name'] == col_name][0]
            col_preds = self.models[col_name].predict(data)
            ypred[col_name] = {}
            if col_name in self.ord_encs:
                ypred[col_name]['class_distribution'] = list(col_preds)
                ypred[col_name]['class_labels'] = {i: cls for i, cls in enumerate(self.all_classes)}
                col_preds = self.ord_encs[col_name].inverse_transform(np.argmax(col_preds, axis=1).reshape(-1, 1)).flatten()
            if col_config.get('encoder_attrs', False):
                if col_config['encoder_attrs'].get('positive_domain', False):
                    col_preds = col_preds.clip(0)
            ypred[col_name]['predictions'] = list(col_preds)

        return ypred



import torch

from lightwood.encoders.encoder_base import BaseEncoder
from lightwood.constants.lightwood import COLUMN_DATA_TYPES
from lightwood.encoders.time_series.helpers.common import MinMaxNormalizer, CatNormalizer


class TimeSeriesPlainEncoder(BaseEncoder):
    def __init__(self, is_target=False):
        """
        Fits a normalizer for a time series previous historical data.
        When encoding, it returns a normalized window of previous data.
        """
        super().__init__(is_target)
        self.original_type = None
        self._normalizer = None

    def prepare(self, priming_data):
        if self._prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        if self.original_type == COLUMN_DATA_TYPES.CATEGORICAL:
            self._normalizer = CatNormalizer(encoder_class='ordinal')
        else:
            self._normalizer = MinMaxNormalizer()

        self._normalizer.prepare(priming_data)
        self._prepared = True

    def encode(self, column_data):
        if not self._prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')
        data = torch.cat([self._normalizer.encode(column_data)], dim=-1)
        data[torch.isnan(data)] = 0.0
        data[torch.isinf(data)] = 0.0
        return data
